import pytest

import f3d


def test_properties():
    engine = f3d.Engine.create(True)
    camera = engine.window.camera

    pos = 1, 2, 3
    foc = 1, 22, 3
    up = 0, 0, 1
    angle = 32

    camera.position = pos
    camera.focal_point = foc
    camera.view_up = up
    camera.view_angle = angle
    assert camera.position == pos
    assert camera.focal_point == foc
    assert camera.view_up == up
    assert camera.view_angle == angle


def test_get_state():
    engine = f3d.Engine.create(True)
    camera = engine.window.camera

    pos = 1, 2, 3
    foc = 1, 22, 3
    up = 0, 0, 1
    angle = 32

    camera.position = pos
    camera.focal_point = foc
    camera.view_up = up
    camera.view_angle = angle

    assert camera.state.position == pos
    assert camera.state.focal_point == foc
    assert camera.state.view_up == up
    assert camera.state.view_angle == angle


def test_set_state():
    engine = f3d.Engine.create(True)
    camera = engine.window.camera

    state = f3d.CameraState((1, 2, 3), (1, 22, 3), (0, 0, 1), 32)
    camera.state = state

    assert camera.position == state.position
    assert camera.focal_point == state.focal_point
    assert camera.view_up == state.view_up
    assert camera.view_angle == state.view_angle


def test_default_state():
    new_state = f3d.CameraState()
    assert new_state.position == (0, 0, 1)
    assert new_state.focal_point == (0, 0, 0)
    assert new_state.view_up == (0, 1, 0)
    assert new_state.view_angle == 30


@pytest.mark.xfail(reason="CameraState equality not implemented")
def test_state_compare():
    state1 = f3d.CameraState((1, 2, 3), (1, 22, 3), (0, 0, 1), 32)
    state2 = f3d.CameraState((1, 2, 3), (1, 22, 3), (0, 0, 1), 32)
    assert state1 == state2

    state3 = f3d.CameraState((1, 2, 3), (1, 22, 3), (0, 0, 1), 25)
    assert state1 != state3


def test_moves():
    engine = f3d.Engine.create(True)
    camera = engine.window.camera

    camera.dolly(10)
    angle = 30
    camera.roll(angle)
    camera.azimuth(angle)
    camera.yaw(angle)
    camera.elevation(angle)
    camera.pitch(angle)


def test_pan():
    engine = f3d.Engine.create(True)
    camera = engine.window.camera

    camera.state = f3d.CameraState((1, 2, 3), (1, 2, 13), (0, 1, 0), 40)
    camera.pan(1, 2)
    assert camera.state.position == (0, 4, 3)
    assert camera.state.focal_point == (0, 4, 13)

    camera.state = f3d.CameraState((1, 2, 3), (1, -2, 3), (0, 0, 1), 40)
    camera.pan(3, 4, 5)
    assert camera.state.position == (-2, -3, 7)
    assert camera.state.focal_point == (-2, -7, 7)


def test_resets():
    engine = f3d.Engine.create(True)
    camera = engine.window.camera
    camera.set_current_as_default()
    camera.reset_to_bounds()
    camera.reset_to_default()



import pytest

import f3d


def print_fn(args: list[str]):
    print(args)


def doc_fn():
    return "dummyDoc", "dummyValue"


def test_command(capfd: pytest.CaptureFixture[str]):
    engine = f3d.Engine.create(True)
    inter = engine.interactor

    # Check default commands can be removed
    actions = inter.get_command_actions()
    for action in actions:
        inter.remove_command(action)
    assert len(inter.get_command_actions()) == 0

    # Check a command can be triggered
    inter.add_command("my_cmd", print_fn)
    inter.trigger_command("my_cmd arg1 arg2")
    inter.remove_command("my_cmd")
    out, _err = capfd.readouterr()
    assert out == "['arg1', 'arg2']\n"

    # Smoke test
    inter.init_commands()


def test_binding():
    engine = f3d.Engine.create(True)
    inter = engine.interactor

    # Check default interactions can be removed
    groups = inter.get_bind_groups()
    for group in groups:
        binds = inter.get_binds_for_group(group)
        for bind in binds:
            inter.remove_binding(bind)
    assert len(inter.get_bind_groups()) == 0
    assert len(inter.get_binds()) == 0

    # Smoke test
    inter.add_binding(
        f3d.InteractionBind(f3d.InteractionBind.ModifierKeys.ANY, "P"),
        "dummy command",
        "DummyGroup",
        doc_fn,
    )
    inter.add_binding(
        f3d.InteractionBind(f3d.InteractionBind.ModifierKeys.NONE, "P"),
        "dummy command",
        "DummyGroup",
        doc_fn,
    )
    inter.add_binding(
        f3d.InteractionBind(f3d.InteractionBind.ModifierKeys.CTRL, "P"),
        "dummy command",
        "DummyGroup",
        doc_fn,
    )
    inter.add_binding(
        f3d.InteractionBind(f3d.InteractionBind.ModifierKeys.SHIFT, "P"),
        "dummy command",
        "DummyGroup",
        doc_fn,
    )
    inter.add_binding(
        f3d.InteractionBind(f3d.InteractionBind.ModifierKeys.CTRL_SHIFT, "P"),
        ["dummy command", "dummy command"],
        "DummyGroup",
        doc_fn,
    )
    assert len(inter.get_bind_groups()) == 1
    assert len(inter.get_binds()) == 5

    inter.init_bindings()



from pathlib import Path
import tempfile

import f3d


def test_compare_with_file():
    testing_dir = Path(__file__).parent.parent.parent / "testing"
    dataset = testing_dir / "data/cow.vtp"
    reference = testing_dir / "baselines/TestPythonCompareWithFile.png"
    output = Path(tempfile.gettempdir()) / "TestPythonCompareWithFile.png"

    engine = f3d.Engine.create(True)
    engine.window.size = 300, 300

    # verify the size is properly set
    assert engine.window.width == 300
    assert engine.window.height == 300

    engine.scene.add(dataset)

    img = engine.window.render_to_image()
    img.save(output)

    assert img.compare(f3d.Image(reference)) <= 0.05



import subprocess
import sys


def test_default_level():
    assert (
        run_python(
            "from f3d import Log",
            "Log.set_use_coloring(False)",
            "Log.print(Log.DEBUG, 'debug')",
            "Log.print(Log.INFO, 'info')",
        )
        == "info\n"
    )


def test_debug():
    assert (
        run_python(
            "from f3d import Log",
            "Log.set_use_coloring(False)",
            "Log.set_verbose_level(Log.DEBUG)",
            "Log.print(Log.DEBUG, 'debug')",
        )
        == "debug\n"
    )


def test_coloring():
    assert (
        run_python(
            "from f3d import Log",
            "Log.set_use_coloring(True)",
            "Log.print(Log.INFO, 'info')",
        )
        == "info\x1b[0m\n"
    )


def run_python(*statements: str):
    return subprocess.check_output(
        [sys.executable, "-c", "; ".join(statements)],
        text=True,
    )



import re
import subprocess
import sys
from argparse import ArgumentParser
from contextlib import contextmanager
from difflib import unified_diff
from pathlib import Path
from tempfile import gettempdir
from typing import Iterable


def main():
    argparser = ArgumentParser()
    argparser.add_argument(
        "-o",
        "--into",
        help="output directory for the post-processed stubs (default: %(default)s)",
        default=f"{gettempdir()}/stubs",
    )
    args = argparser.parse_args()

    stubs = run_pybind11_stubgen(Path(args.into))
    if diff := postprocess_generated_stubs(stubs):
        print("\n".join(diff))


def run_pybind11_stubgen(out_dir: Path, module: str = "f3d"):
    stubgen_cmd = (
        # use current python interpreter to run stubs generation for the `f3d` module
        *(sys.executable, "-m", "pybind11_stubgen", module),
        # fix enum for default values in `Image.save()` and `Image.save_buffer()`
        *("--enum-class-locations", "SaveFormat:Image"),
        # ignore `f3d.vector3_t` and `f3d.point3_t` as we dont actually map them
        # but let them auto-convert from and to `tuple[float, float, float]`
        # (all occurrences will be postprocessed later)
        *("--ignore-unresolved-names", r"f3d\.(vector3_t|point3_t)"),
        # output directory so we can retrieve and post process
        *("--output-dir", out_dir),
    )
    with retrieve_changed_files(out_dir, f"{module}/**/*.pyi") as changed_files:
        subprocess.check_call(stubgen_cmd)
    return changed_files


def postprocess_generated_stubs(filenames: Iterable[Path]):
    replacements = [
        (
            # change `point3_t` and `vector3_t` parameter annotations and return types
            # to `tuple[float, float, float]`
            r"((:|->)\s*)f3d\.(vector3_t|point3_t)",
            r"\1tuple[float, float, float]",
        ),
        (
            # add missing template parameter to raw `os.PathLike` (`os.PathLike[str]`)
            r"(PathLike)(?!\[)",
            r"\1[str]",
        ),
        (
            # remove `_pybind11_conduit_v1_` static methods
            r"^\s+@staticmethod\s+def _pybind11_conduit_v1_\(\*args, *\*\*kwargs\):\s*\.\.\.[\n\r]",
            "",
        ),
    ]

    diff: list[str] = []

    for filename in filenames:
        processed = original = filename.read_text()
        for pattern, repl in replacements:
            processed = re.sub(pattern, repl, processed, flags=re.MULTILINE)
        filename.write_text(processed)

        diff += unified_diff(
            original.splitlines(),
            processed.splitlines(),
            str(filename),
            str(filename),
            n=1,
            lineterm="",
        )

    return diff


@contextmanager
def retrieve_changed_files(directory: Path, files_glob: str):
    mtimes = {f: f.stat().st_mtime for f in directory.glob(files_glob)}
    changed_files: list[Path] = []
    yield changed_files
    changed_files += (
        f for f in directory.glob(files_glob) if f.stat().st_mtime > mtimes.get(f, 0)
    )


if __name__ == "__main__":
    main()



from argparse import ArgumentParser

import f3d

if __name__ == "__main__":
    argparser = ArgumentParser()
    argparser.add_argument("image1")
    argparser.add_argument("image2")

    args = argparser.parse_args()

    img1 = f3d.Image(args.image1)
    img2 = f3d.Image(args.image2)

    error = img1.compare(img2)
    if error > 0.05:
        print(f"Images are different ({error=})")
    else:
        print(f"Images are identical ({error=})")



import os
from pathlib import Path
import tempfile
import pytest

import f3d


@pytest.fixture
def f3d_engine():
    engine = f3d.Engine.create(True)
    engine.window.size = 300, 200
    return engine


def test_render_rgb(f3d_engine: f3d.Engine):
    window = f3d_engine.window

    img = window.render_to_image()
    assert img.width == window.width
    assert img.height == window.height
    assert img.channel_count == 3

    data = img.content
    assert isinstance(data, (bytes, bytearray))
    assert len(data) == img.channel_count * img.width * img.height


def test_render_rgba(f3d_engine: f3d.Engine):
    window = f3d_engine.window

    img = window.render_to_image(True)
    assert img.width == window.width
    assert img.height == window.height
    assert img.channel_count == 4
    data = img.content
    assert isinstance(data, (bytes, bytearray))
    assert len(data) == img.channel_count * img.width * img.height


def test_set_data(f3d_engine: f3d.Engine):
    img = f3d_engine.window.render_to_image()
    data = img.content[:]
    img.content = data
    assert img.content == data


def test_set_wrong_data(f3d_engine: f3d.Engine):
    img = f3d_engine.window.render_to_image()
    with pytest.raises(ValueError):
        img.content = img.content[:-1]


def test_save(f3d_engine: f3d.Engine):
    img = f3d_engine.window.render_to_image()
    fn = Path(tempfile.gettempdir()) / "TestPythonSaveFile.bmp"

    img.save(fn, f3d.Image.SaveFormat.BMP)
    assert os.path.isfile(fn)


def test_save_buffer(f3d_engine: f3d.Engine):
    img = f3d_engine.window.render_to_image(True)
    buffer = img.save_buffer(f3d.Image.SaveFormat.PNG)
    assert buffer.startswith(b"\x89PNG")
    assert img._repr_png_() == buffer


def test_formats(f3d_engine: f3d.Engine):
    formats = f3d.Image.supported_formats()
    assert ".png" in formats


def test_normalized_pixel(f3d_engine: f3d.Engine):
    img = f3d_engine.window.render_to_image()
    assert img.normalized_pixel((0, 0)) == [0.2, 0.2, 0.2]


@pytest.mark.parametrize(
    "img_filename",
    [
        "toTerminalText-rgb.png",
        "toTerminalText-rgba.png",
    ],
)
def test_to_terminal_text(img_filename: str):
    testing_data_dir = Path(__file__).parent.parent.parent / "testing/data"
    image_path = testing_data_dir / img_filename
    text_path = image_path.with_suffix(".txt")
    assert (
        f3d.Image(image_path).to_terminal_text()
        == open(text_path, encoding="utf8").read()
    )


def test_metadata():
    image = f3d.Image(4, 2, 3, f3d.Image.ChannelType.BYTE)
    image.set_metadata("foo", "bar")
    image.set_metadata("hello", "world")
    assert image.get_metadata("foo") == "bar" and image.get_metadata("hello") == "world"

    with pytest.raises(KeyError):
        image.get_metadata("baz")

    assert set(image.all_metadata()) == set(["foo", "hello"])



import f3d

from pathlib import Path


def test_text_distance():
    assert f3d.Utils.text_distance("modle", "model") == 2


def test_collapse_path():
    assert f3d.Utils.collapse_path("/folder/../file.ext", ".") == Path("/file.ext")



from functools import partial

import f3d


def stop_fn(inter: f3d.Interactor):
    inter.stop()


def test_start_stop():
    engine = f3d.Engine.create(True)
    inter = engine.interactor
    inter.start(1 / 30, partial(stop_fn, inter))


def test_request_render():
    engine = f3d.Engine.create(True)
    inter = engine.interactor
    inter.request_render()
    inter.start(1 / 30, partial(stop_fn, inter))
